import itertools
import sys
import typing
import warnings
from ast import *

operator_map: dict[type[operator], str] = {
    Add: "+",
    BitAnd: "&",
    BitOr: "|",
    BitXor: "^",
    Div: "/",
    FloorDiv: "//",
    LShift: "<<",
    Mod: "%",
    Mult: "*",
    MatMult: "@",
    Pow: "**",
    RShift: ">>",
    Sub: "-",
}

boolop_map: dict[type[boolop], str] = {
    And: "and",
    Or: "or",
}

unaryop_map: dict[type[unaryop], str] = {
    UAdd: "+",
    USub: "-",
    Invert: "~",
    Not: "not ",
}

cmpop_map: dict[type[cmpop], str] = {
    Eq: "==",
    Gt: ">",
    GtE: ">=",
    In: " in ",
    Is: " is ",
    IsNot: " is not ",
    Lt: "<",
    LtE: "<=",
    NotEq: "!=",
    NotIn: " not in ",
}
enum = itertools.count()
PERC_NAME = next(enum)
PERC_ATTR = next(enum)
PREC_AWAIT = next(enum)
PREC_POW = next(enum)
PREC_INV = next(enum)
PREC_MULT = next(enum)
PREC_ADD = next(enum)
PREC_SHIFT = next(enum)
PREC_BITAND = next(enum)
PREC_BITXOR = next(enum)
PREC_BITOR = next(enum)
PREC_COMP_ARGS = next(enum)
PREC_COMPARE = next(enum)
PREC_NOT = next(enum)
PREC_AND = next(enum)
PREC_OR = next(enum)
PREC_IFEXP_LEFT = next(enum)
PREC_IFEXP = next(enum)
PREC_FORMAT_EXPR = next(enum)
PREC_LAMBDA = next(enum)
PREC_EXPR = next(enum)
PREC_NAMEDEXPR = next(enum)
PREC_CALLARGS = next(enum)
PREC_GENEXPR = next(enum)
PREC_ONECALLARG = next(enum)
PREC_YIELD = next(enum)
INF = float("inf")


def get_BinOp_precedence(op):
    op_type = type(op)
    if op_type is Pow:
        return PREC_POW
    if op_type in [Mult, MatMult, Div, FloorDiv, Mod]:
        return PREC_MULT
    if op_type in [Add, Sub]:
        return PREC_ADD
    if op_type in [LShift, RShift]:
        return PREC_SHIFT
    if op_type is BitAnd:
        return PREC_BITAND
    if op_type is BitXor:
        return PREC_BITXOR
    if op_type is BitOr:
        return PREC_BITOR
    return INF


def get_BoolOp_precedence(op):
    op_type = type(op)
    if op_type is And:
        return PREC_AND
    if op_type is Or:
        return PREC_OR
    return INF


def get_UnaryOp_precedence(op):
    op_type = type(op)
    if op_type in [UAdd, USub, Invert]:
        return PREC_INV
    if op_type is Not:
        return PREC_NOT
    return INF


ast_prec_map: dict[type[AST] : int] = {
    Constant: PERC_NAME,
    JoinedStr: PERC_NAME,
    FormattedValue: PERC_NAME,
    List: PERC_NAME,
    ListComp: PERC_NAME,
    Tuple: PERC_NAME,
    Dict: PERC_NAME,
    DictComp: PERC_NAME,
    Set: PERC_NAME,
    SetComp: PERC_NAME,
    Name: PERC_NAME,
    Starred: PERC_NAME,
    Attribute: PERC_ATTR,
    Subscript: PERC_ATTR,
    Call: PERC_ATTR,
    Await: PREC_AWAIT,
    Compare: PREC_COMPARE,
    IfExp: PREC_IFEXP,
    Lambda: PREC_LAMBDA,
    keyword: PREC_EXPR,
    Slice: PREC_EXPR,
    NamedExpr: PREC_NAMEDEXPR,
    GeneratorExp: PREC_GENEXPR,
    Yield: PREC_YIELD,
    YieldFrom: PREC_YIELD,
}


def get_precedence(node):
    _type = type(node)
    try:
        return ast_prec_map[type(node)]
    except KeyError:
        pass

    if _type is BinOp:
        return get_BinOp_precedence(node.op)
    if _type is UnaryOp:
        return get_UnaryOp_precedence(node.op)
    if _type is BoolOp:
        return get_BoolOp_precedence(node.op)
    warnings.warn(f"Unknown node precedence of '{type(node).__name__}'", RuntimeWarning)
    return INF


def unparse_generic(node: AST):
    warnings.warn(f"Unknown node type '{type(node).__name__}'", RuntimeWarning)
    return ""
    yield


def unparse_Name(node: Name):
    return node.id
    yield


def get_unescaped_str(string: str, qm: str):
    out = []
    for i in string:
        if i == qm:
            out.append(f"\\{qm}")
        elif ord(i) > 255:
            out.append(i)
        else:
            out.append(ascii(i)[1:-1])
    return "".join(out)


def unparse_Constant(node: Constant, qm):
    if node.value is ...:
        return "..."
    if isinstance(node.value, str):
        value = get_unescaped_str(node.value, qm)
        return f"{qm}{value}{qm}"
    return repr(node.value)
    yield


def _unparse_JoinedStr(node: JoinedStr, qm):
    contents = []
    for v in node.values:
        if isinstance(v, Constant):
            assert isinstance(v.value, str)
            s = get_unescaped_str(v.value, qm)
            s = s.replace("{", "{{").replace("}", "}}")
            contents.append(s)
        elif isinstance(v, FormattedValue):
            contents.append((yield PREC_FORMAT_EXPR, v))
    return "".join(contents)


def unparse_JoinedStr(node: JoinedStr, qm):
    contents = yield from _unparse_JoinedStr(node, qm)
    if sys.version_info < (3, 12) and "\\" in contents:
        raise RuntimeError("Back slash is included in a f-string")
    return f"f{qm}{contents}{qm}"


def unparse_FormattedValue(node: FormattedValue, qm):
    value = yield PREC_FORMAT_EXPR, node.value
    format_spec = ""
    if node.format_spec is not None:
        assert isinstance(node.format_spec, JoinedStr)
        format_spec = yield from _unparse_JoinedStr(node.format_spec, qm)
        format_spec = ":" + format_spec
    return f"{{{value}{format_spec}}}"


def unparse_Starred(node: Starred):
    precedence = PERC_NAME
    value = yield precedence, node.value
    return f"*{value}"
    yield


def unparse_Attribute(node: Attribute):
    precedence = PERC_NAME
    value = yield precedence, node.value
    return f"{value}.{node.attr}"


def unparse_Subscript(node: Subscript):
    precedence = PERC_NAME
    value = yield precedence, node.value
    _slice = yield INF, node.slice
    return f"{value}[{_slice}]"


def unparse_Slice(node: Slice):
    precedence = PREC_EXPR
    upper, lower, step = "", "", ""
    if node.upper is not None:
        upper = yield precedence, node.upper
    if node.lower is not None:
        lower = yield precedence, node.lower
    if node.step is not None:
        step = yield precedence, node.step
    return f"{lower}:{upper}:{step}"


def unparse_keyword(node: keyword):
    precedence = PREC_EXPR
    value = yield precedence, node.value
    if node.arg is None:
        return f"**{value}"
    return f"{node.arg}={value}"


def unparse_Call(node: Call):
    func = yield PERC_ATTR, node.func
    if len(node.args) == 1 and len(node.keywords) == 0:
        _arg = yield PREC_ONECALLARG, node.args[0]
        return f"{func}({_arg})"
    args_list = []
    for _arg in itertools.chain(node.args, node.keywords):
        args_list.append((yield PREC_CALLARGS, _arg))
    _args = ",".join(args_list)
    return f"{func}({_args})"


def unparse_BinOp(node: BinOp):
    precedence = get_BinOp_precedence(node.op)
    op = operator_map[type(node.op)]
    left = yield precedence, node.left
    right = yield precedence, node.right
    return f"{left}{op}{right}"


def unparse_BoolOp(node: BoolOp):
    precedence = get_BoolOp_precedence(node.op)
    op = boolop_map[type(node.op)]
    values = []
    for v in node.values:
        values.append((yield precedence, v))
    return f" {op} ".join(values)


def unparse_UnaryOp(node: UnaryOp):
    precedence = get_UnaryOp_precedence(node.op)
    op = unaryop_map[type(node.op)]
    operand = yield precedence, node.operand
    return f"{op}{operand}"


def unparse_List(node: List):
    elts = []
    for item in node.elts:
        elts.append((yield PREC_CALLARGS, item))
    return f"[{','.join(elts)}]"


def unparse_Set(node: Set):
    elts = []
    for item in node.elts:
        elts.append((yield PREC_CALLARGS, item))
    return f"{{{','.join(elts)}}}"


def unparse_Dict(node: Dict):
    item = []
    for k, v in zip(node.keys, node.values):
        key = yield PREC_EXPR, k
        value = yield PREC_EXPR, v
        item.append(f"{key}:{value}")
    return f"{{{','.join(item)}}}"


def unparse_Tuple(node: Tuple):
    elts = []
    for item in node.elts:
        elts.append((yield PREC_CALLARGS, item))
    if len(elts) == 1:
        return f"({elts[0]},)"
    return f"({','.join(elts)})"


def unparse_Compare(node: Compare):
    ops = (cmpop_map[type(op)] for op in node.ops)
    left = yield PREC_COMP_ARGS, node.left
    comparators = []
    for comparator in node.comparators:
        comparators.append((yield PREC_COMP_ARGS, comparator))
    right = "".join(op + comp for op, comp in zip(ops, comparators))
    return f"{left}{right}"


def unparse_NamedExpr(node: NamedExpr):
    precedence = PREC_NAMEDEXPR
    value = yield precedence, node.value
    return f"{node.target.id}:={value}"


def unparse_Lambda(node: Lambda):
    body = yield PREC_LAMBDA, node.body
    arg_def_list = []

    # handle posonly args
    for posonly in node.args.posonlyargs:
        arg_def_list.append(posonly.arg)

    # handle args
    for _arg in node.args.args:
        arg_def_list.append(_arg.arg)
    ind = len(arg_def_list)
    for default in reversed(node.args.defaults):
        ind -= 1
        if default is not None:
            arg_def_list[ind] += f"={yield PREC_EXPR,default}"

    if node.args.posonlyargs:
        arg_def_list.insert(len(node.args.posonlyargs), "/")

    # handle vararg
    if node.args.vararg:
        arg_def_list.append(f"*{node.args.vararg.arg}")
    elif node.args.kwonlyargs:
        arg_def_list.append("*")

    # handle kwonly args
    kw_list = []
    for kwonly in node.args.kwonlyargs:
        kw_list.append(kwonly.arg)
    for ind, default in enumerate(node.args.kw_defaults):
        if default is not None:
            kw_list[ind] += f"={yield PREC_EXPR,default}"
    arg_def_list.extend(kw_list)

    # handle kwarg
    if node.args.kwarg:
        arg_def_list.append(f"**{node.args.kwarg.arg}")

    arg_def = ",".join(arg_def_list)
    if arg_def:
        arg_def = " " + arg_def
    return f"lambda{arg_def}:{body}"


def _unparse_comprehensions(generators):
    generator_list = []
    for gen in generators:
        _async = "" if not gen.is_async else "async "
        _iter = yield PREC_EXPR, gen.iter
        target = yield PREC_EXPR, gen.target
        if_list = []
        for test in gen.ifs:
            if_list.append((yield PREC_EXPR, test))
        ifs = " if ".join(if_list)
        if if_list:
            ifs = " if " + ifs
        generator_list.append(f"{_async}for {target} in {_iter}{ifs}")
    return " ".join(generator_list)


def unparse_ListComp(node: ListComp):
    elt = yield PREC_CALLARGS, node.elt
    generators = yield from _unparse_comprehensions(node.generators)
    return f"[{elt} {generators}]"


def unparse_GeneratorExp(node: GeneratorExp):
    elt = yield PREC_CALLARGS, node.elt
    generators = yield from _unparse_comprehensions(node.generators)
    return f"{elt} {generators}"


def unparse_SetComp(node: SetComp):
    elt = yield PREC_CALLARGS, node.elt
    generators = yield from _unparse_comprehensions(node.generators)
    return f"{{{elt} {generators}}}"


def unparse_DictComp(node: DictComp):
    key = yield PREC_EXPR, node.key
    value = yield PREC_EXPR, node.value
    generators = yield from _unparse_comprehensions(node.generators)
    return f"{{{key}:{value} {generators}}}"


def unparse_IfExp(node: IfExp):
    body = yield PREC_IFEXP_LEFT, node.body
    test = yield PREC_IFEXP_LEFT, node.test
    orelse = yield PREC_EXPR, node.orelse
    return f"{body} if {test} else {orelse}"


def unparse_Yield(node: Yield):
    if node.value is None:
        return "yield"
    value = yield PREC_EXPR, node.value
    return f"yield {value}"


def unparse_YieldFrom(node: YieldFrom):
    value = yield PREC_EXPR, node.value
    return f"yield from {value}"


class _Node:
    gen_map: dict[type[AST], typing.Callable] = {
        Name: unparse_Name,
        Call: unparse_Call,
        Constant: unparse_Constant,
        JoinedStr: unparse_JoinedStr,
        FormattedValue: unparse_FormattedValue,
        Starred: unparse_Starred,
        keyword: unparse_keyword,
        BinOp: unparse_BinOp,
        BoolOp: unparse_BoolOp,
        UnaryOp: unparse_UnaryOp,
        List: unparse_List,
        Tuple: unparse_Tuple,
        Set: unparse_Set,
        Dict: unparse_Dict,
        Compare: unparse_Compare,
        Attribute: unparse_Attribute,
        Subscript: unparse_Subscript,
        Slice: unparse_Slice,
        NamedExpr: unparse_NamedExpr,
        Lambda: unparse_Lambda,
        ListComp: unparse_ListComp,
        SetComp: unparse_SetComp,
        DictComp: unparse_DictComp,
        GeneratorExp: unparse_GeneratorExp,
        IfExp: unparse_IfExp,
        Yield: unparse_Yield,
        YieldFrom: unparse_YieldFrom,
    }

    def __init__(self, outer_precedence, node: AST, outer_str_qm: str):
        self.outer_precedence = outer_precedence
        self.precedence = get_precedence(node)
        gen_func = self.gen_map.get(type(node), unparse_generic)

        if gen_func in [unparse_Constant, unparse_JoinedStr]:
            if outer_str_qm == "'":
                self.qm = '"'
            elif outer_str_qm == '"':
                self.qm = "'"
            self.gen = gen_func(node, self.qm)
        elif gen_func is unparse_FormattedValue:
            self.qm = outer_str_qm
            self.gen = gen_func(node, self.qm)
        else:
            self.qm = outer_str_qm
            self.gen = gen_func(node)


def unparse(node: expr):
    stack: list[_Node] = []
    stack.append(_Node(PREC_EXPR, node, '"'))
    converted: str | None = None
    unconverted: tuple[int, AST] | None = None
    while stack:
        try:
            unconverted = stack[-1].gen.send(converted)
            stack.append(_Node(*unconverted, stack[-1].qm))  # type: ignore
            converted = None
        except StopIteration as err:
            converted = err.value
            inner_node = stack.pop()
            if inner_node.precedence > inner_node.outer_precedence:
                converted = f"({converted})"

    return converted



import argparse

import oneliner
import oneliner.version

parser = argparse.ArgumentParser(
    description="Convert python scripts into oneliner expression."
)

parser.add_argument(
    "input_filename",
    type=str,
    help="The filename of the python script to be converted",
)

ver = oneliner.version.oneliner_version
dev = "-dev" if oneliner.version.dev else ""
parser.add_argument(
    "-v",
    "--version",
    action="version",
    version=f"Oneliner-Py-{ver[0]}.{ver[1]}.{ver[2]}{dev}",
)

parser.add_argument(
    "-o",
    "--output",
    type=str,
    help="The output filename. If this argument is not specified, "
    "Oneliner-Py will print the result to the screen.",
)

parser.add_argument(
    "--unparser",
    type=str,
    default="ast.unparse",
    choices=["ast.unparse", "oneliner"],
)

args = parser.parse_args()

with open(args.input_filename, "r", encoding="utf8") as infile:
    script = infile.read()

if args.unparser == "ast.unparse":
    converted = oneliner.convert_code_string(script, use_new_unparser=False)
else:
    converted = oneliner.convert_code_string(script, use_new_unparser=True)

if args.output is not None:
    with open(args.output, "w", encoding="utf8") as outfile:
        outfile.write(converted)
else:
    print(converted)



from oneliner.convert import OnelinerConvertor
from oneliner.expr_unparse import unparse

__all__ = ["OnelinerConvertor", "convert_code_string"]

import ast
import symtable


def convert_code_string(code: str, filename="<string>", use_new_unparser=False):
    ast_root = ast.parse(code, filename, "exec")
    symtable_root = symtable.symtable(code, filename, "exec")
    out = OnelinerConvertor().cvt(ast_root, symtable_root)
    if use_new_unparser:
        return unparse(out)
    else:
        return ast.unparse(out).replace("\n", "")



import random
import typing
from ast import AST, Call, Constant, List, Load, Name, Slice


def unique_id() -> str:
    return "".join(random.choices("abcdefghijklmnopqrstuvwxyz", k=10))


def convert_slice(_slice: Slice) -> Call:
    """
    Convert slice expt to a call of slice function
    to prevent invalid syntax like `__setitem__(0:10:2, value)`
    """
    _slice_value = lambda v: Constant(None) if v is None else v
    return Call(
        func=Name(id="slice", ctx=Load()),
        args=[
            _slice_value(_slice.lower),
            _slice_value(_slice.upper),
            _slice_value(_slice.step),
        ],
        keywords=[],
    )


def list_wrapper(nodes: list[AST]) -> AST:
    """
    Wrap a list of nodes as an `ast.List` node
    """
    if len(nodes) == 0:
        return Constant(value=...)
    if len(nodes) == 1:
        return nodes[0]
    return List(elts=nodes, ctx=Load())


def never_call(*args, **kwargs) -> typing.NoReturn:
    raise RuntimeError("this function should never be called")  # pragma: no cover


def ast_debug_info(node: AST):
    return f"At line {node.lineno}, col {node.col_offset}: "



import ast
import symtable

import oneliner.utils as utils
from oneliner.namespaces import *
from oneliner.pending_nodes import *


class OnelinerConvertor:
    def __init__(self):
        self.pending_node_stack: list[PendingNode] = []
        self.nsp_stack: list[Namespace] = []

    _pending_map: dict[type[ast.AST], type[PendingNode]] = {
        ast.Module: PendingModule,
        ast.Expr: PendingExpr,
        ast.If: PendingIf,
        ast.While: PendingWhile,
        ast.For: PendingFor,
        ast.Break: PendingBreak,
        ast.Continue: PeindingContinue,
        ast.Pass: PendingPass,
        ast.Assign: PendingAssign,
        ast.AnnAssign: PendingAssign,
        ast.AugAssign: PendingAugAssign,
        ast.FunctionDef: PendingFunctionDef,
        ast.Return: PendingReturn,
        ast.Global: PendingGlobal,
        ast.Nonlocal: PendingNonlocal,
        ast.ClassDef: PendingClassDef,
        ast.Import: PendingImport,
        ast.ImportFrom: PendingImportFrom,
    }

    def get_pending_node(self, node: ast.AST) -> PendingNode:
        try:
            return self._pending_map[type(node)](
                node,
                nsp=self.nsp_stack[-1],
                nsp_global=self.nsp_global,
            )
        except KeyError as err:
            raise RuntimeError(
                utils.ast_debug_info(node)
                + f"Unable to convert node '{type(node).__name__}'"
            ) from err

    def pending_top(self) -> PendingNode:
        """Get the stack top of self.pending_node_stack"""
        return self.pending_node_stack[-1]

    def cvt(self, ast_root: ast.Module, symtable_root: symtable.SymbolTable) -> ast.AST:
        self.nsp_global = generate_nsp(symtable_root)
        self.nsp_stack.append(self.nsp_global)
        unconverted = ast_root
        result_nodes = None
        while True:
            pending_node = self.get_pending_node(unconverted)
            self.pending_node_stack.append(pending_node)
            if pending_node.has_internal_namespace:
                self.nsp_stack.append(pending_node.get_internal_namespace())
            unconverted = None
            while unconverted is None:
                try:
                    # try to get unconverted node
                    if result_nodes is None:
                        unconverted = next(self.pending_top().iter_node)
                    else:
                        unconverted = self.pending_top().iter_node.send(result_nodes)
                        result_nodes = None
                except StopIteration:
                    # no more unconverted node, pending node is complete
                    complete_node = self.pending_node_stack.pop()
                    if complete_node.has_internal_namespace:
                        self.nsp_stack.pop()
                    result_nodes = complete_node.get_result()

                    if len(self.pending_node_stack) == 0:
                        assert len(self.nsp_stack) == 1
                        return utils.list_wrapper(result_nodes)



import typing
from ast import *

import oneliner.utils as utils
from oneliner.expr_transform import expr_transf
from oneliner.namespaces import (
    Namespace,
    NamespaceClass,
    NamespaceFunction,
    NamespaceGlobal,
)
from oneliner.reserved_identifiers import *

__all__ = [
    "PendingNode",
    "PendingModule",
    "PendingExpr",
    "PendingIf",
    "PendingWhile",
    "PendingFor",
    "PendingBreak",
    "PeindingContinue",
    "PendingPass",
    "PendingAssign",
    "PendingAugAssign",
    "PendingFunctionDef",
    "PendingReturn",
    "PendingGlobal",
    "PendingNonlocal",
    "PendingClassDef",
    "PendingImport",
    "PendingImportFrom",
]


class PendingNode:
    def __init__(self, node: AST, nsp: Namespace, nsp_global: NamespaceGlobal):
        self.iter_node = self._iter_nodes()
        self.nsp = nsp
        self.nsp_global = nsp_global

    def get_result(self) -> list[AST]:
        raise NotImplementedError()  # pragma: no cover

    def _iter_nodes(self) -> typing.Generator[AST, list[AST], None]:
        return None
        yield AST()

    has_internal_namespace: bool = False

    def get_internal_namespace(self) -> Namespace:
        raise NotImplementedError()  # pragma: no cover


class PendingModule(PendingNode):
    def __init__(self, node: Module, nsp: Namespace, nsp_global: NamespaceGlobal):
        super().__init__(node, nsp, nsp_global)
        self.node = node
        self.converted_body = []

    def _iter_nodes(self) -> typing.Generator[AST, list[AST], None]:
        for node in self.node.body:
            self.converted_body.extend((yield node))

    def _insert_import_lib(self, libname, asname):
        import_itertools_ast = NamedExpr(
            target=Name(id=asname, ctx=Store()),
            value=Call(
                func=Name(id="__import__", ctx=Load()),
                args=[Constant(value=libname)],
                keywords=[],
            ),
        )
        self.converted_body.insert(0, import_itertools_ast)

    def get_result(self) -> list[AST]:
        if self.nsp_global.use_itertools:
            self._insert_import_lib("itertools", "itertools")
        if self.nsp_global.use_importlib:
            self._insert_import_lib("importlib", "importlib")

        if self.nsp_global.use_preset_iter_wrapper:
            from .presets import iter_wrapper_body

            self.converted_body.insert(0, iter_wrapper_body)

        return self.converted_body


class PendingExpr(PendingNode):
    def __init__(self, node: Expr, nsp: Namespace, nsp_global: NamespaceGlobal):
        super().__init__(node, nsp, nsp_global)
        self.node = node

    def get_result(self) -> list[AST]:
        return [expr_transf(self.nsp, self.node.value)]


class _PendingCompoundStmt(PendingNode):
    """
    CompoundStmt:
        PendingIf
        PendingWhile
        PendingFor
        PendingFunctionDef
        PendingClassDef

    They have some common behaviors
    So _PendingCompoundStmt is created
    """

    def _iter_branch(
        self,
        converted_branch: list[AST],
        branch: list[stmt],
        get_interrupt_cnt,
        get_flow_control_expr,
    ) -> typing.Generator[AST, list[AST], None]:
        initial_interrupt_cnt = get_interrupt_cnt()

        converting = []
        stack = [converting]
        for node in branch:
            if get_interrupt_cnt() > initial_interrupt_cnt:
                converting = []
                stack.append(converting)
                initial_interrupt_cnt = get_interrupt_cnt()

            converting.extend((yield node))

            if isinstance(node, (Break, Continue, Return)):
                # remove nodes after an "interrupt operation"
                # since they never run
                break

        while len(stack) > 1:
            # wrap nodes with an "if" to check interrupt at run time
            wrapped = stack.pop()
            stack[-1].append(
                IfExp(
                    test=UnaryOp(op=Not(), operand=get_flow_control_expr()),
                    body=utils.list_wrapper(wrapped),
                    orelse=Constant(value=...),
                )
            )
        converted_branch.extend(stack[0])


class PendingIf(_PendingCompoundStmt):
    def __init__(self, node: If, nsp: Namespace, nsp_global: NamespaceGlobal):
        super().__init__(node, nsp, nsp_global)
        self.node = node

        self.converted_body = []
        self.converted_orelse = []

    def get_result(self) -> list[AST]:
        return [
            IfExp(
                test=expr_transf(self.nsp, self.node.test),
                body=utils.list_wrapper(self.converted_body),
                orelse=utils.list_wrapper(self.converted_orelse),
            )
        ]

    def _iter_nodes(self) -> typing.Generator[AST, list[AST], None]:
        if self.nsp.loop_stack:
            # if inside a loop
            get_interrupt_cnt = lambda: self.nsp.loop_stack[-1].interrupt_cnt
            get_flow_control_expr = self.nsp.loop_stack[-1].get_flow_ctrl_expr
        elif isinstance(self.nsp, NamespaceFunction):
            # if inside a function
            self.nsp: NamespaceFunction  # fix type checker error
            get_interrupt_cnt = lambda: self.nsp.return_cnt
            get_flow_control_expr = self.nsp.get_flow_ctrl_expr
        else:
            get_interrupt_cnt = lambda: 0
            get_flow_control_expr = utils.never_call

        yield from self._iter_branch(
            self.converted_body,
            self.node.body,
            get_interrupt_cnt,
            get_flow_control_expr,
        )
        yield from self._iter_branch(
            self.converted_orelse,
            self.node.orelse,
            get_interrupt_cnt,
            get_flow_control_expr,
        )


class _PendingLoop(_PendingCompoundStmt):
    node: While | For  # Original node
    flow_ctrl_interrupt_expr: Name
    flow_ctrl_interrupt_used: bool
    interrupt_node_bodies: list[
        list[AST]
    ]  # list of bodies of converted return/break/continue nodes
    converted_body: list[AST]  # converted body branch
    converted_orelse: list[AST]  # converted orelse branch
    interrupt_cnt: int  # Continue, Break and Return will increase this counter
    break_cnt: int  # Break and Return will increase this counter

    def get_flow_ctrl_expr(self):
        self.flow_ctrl_interrupt_used = True
        return self.flow_ctrl_interrupt_expr

    def _iter_nodes(self) -> typing.Generator[AST, list[AST], None]:
        yield from self._iter_branch(
            self.converted_body,
            self.node.body,
            lambda: self.interrupt_cnt,
            self.get_flow_ctrl_expr,
        )
        # body finish, pop from loop stack
        self.nsp.loop_stack.pop()

        if self.nsp.loop_stack:
            # if inside a loop
            get_interrupt_cnt = lambda: self.nsp.loop_stack[-1].interrupt_cnt
            get_flow_control_expr = self.nsp.loop_stack[-1].get_flow_ctrl_expr
        elif isinstance(self.nsp, NamespaceFunction):
            # if inside a function
            self.nsp: NamespaceFunction  # fix type checker error
            get_interrupt_cnt = lambda: self.nsp.return_cnt
            get_flow_control_expr = self.nsp.get_flow_ctrl_expr
        else:
            get_interrupt_cnt = lambda: 0
            get_flow_control_expr = utils.never_call

        yield from self._iter_branch(
            self.converted_orelse,
            self.node.orelse,
            get_interrupt_cnt,
            get_flow_control_expr,
        )


class PendingWhile(_PendingLoop):
    node: While

    def __init__(self, node: While, nsp: Namespace, nsp_global: NamespaceGlobal):
        super().__init__(node, nsp, nsp_global)
        self.node = node

        self.converted_body = []
        self.converted_orelse = []

        self.break_cnt = 0
        self.interrupt_cnt = 0

        # flow-control vars
        self.flow_ctrl_break_expr = Name(id=ol_name(OL_BREAK))
        self.flow_ctrl_interrupt_expr = Name(id=ol_name(OL_INTERRUPT))
        self.flow_ctrl_interrupt_used = False
        self.interrupt_node_bodies = []

        self.nsp.loop_stack.append(self)

        self.nsp_global.use_itertools = True

    def get_result(self) -> list[AST]:
        while_loop_final = []

        # if there is break, reset the flow-control var
        # and enable itertools
        if self.break_cnt:
            while_loop_final.append(
                NamedExpr(
                    target=self.flow_ctrl_break_expr,
                    value=Constant(value=False),
                )
            )

        # init the interrupt flow-control var
        if self.flow_ctrl_interrupt_used:
            self.converted_body.insert(
                0,
                NamedExpr(
                    target=self.flow_ctrl_interrupt_expr,
                    value=Constant(value=False),
                ),
            )
            # inject flow_ctrl_interrupt_expr to interrupt nodes
            for interrupt_body in self.interrupt_node_bodies:
                interrupt_body.append(
                    NamedExpr(
                        target=self.flow_ctrl_interrupt_expr,
                        value=Constant(value=True),
                    )
                )

        # add additional check in "test"
        # if there is a break
        if self.break_cnt:
            while_loop_test = BoolOp(
                op=And(),
                values=[
                    UnaryOp(op=Not(), operand=self.flow_ctrl_break_expr),
                    expr_transf(self.nsp, self.node.test),
                ],
            )
        else:
            while_loop_test = expr_transf(self.nsp, self.node.test)

        # "orelse" runs if there's no break
        if self.break_cnt:
            while_loop_orelse = IfExp(
                test=UnaryOp(op=Not(), operand=self.flow_ctrl_break_expr),
                body=utils.list_wrapper(self.converted_orelse),
                orelse=Constant(value=...),
            )
        else:
            while_loop_orelse = utils.list_wrapper(self.converted_orelse)

        # the main body of the oneliner while loop
        while_loop_body = ListComp(
            elt=utils.list_wrapper(self.converted_body),
            generators=[
                comprehension(
                    target=Name(id="_", ctx=Store()),
                    iter=Call(
                        func=Attribute(
                            value=Name(id="itertools", ctx=Load()),
                            attr="takewhile",
                            ctx=Load(),
                        ),
                        args=[
                            Lambda(
                                args=arguments(
                                    posonlyargs=[],
                                    args=[arg(arg="_")],
                                    kwonlyargs=[],
                                    kw_defaults=[],
                                    defaults=[],
                                ),
                                body=while_loop_test,
                            ),
                            Call(
                                func=Attribute(
                                    value=Name(id="itertools", ctx=Load()),
                                    attr="count",
                                    ctx=Load(),
                                ),
                                args=[],
                                keywords=[],
                            ),
                        ],
                        keywords=[],
                    ),
                    ifs=[],
                    is_async=0,
                )
            ],
        )

        # assemble all parts together and return
        while_loop_final.append(while_loop_body)
        # if orelse is not empty, add orelse
        if self.converted_orelse:
            while_loop_final.append(while_loop_orelse)

        return while_loop_final


class PendingFor(_PendingLoop):
    node: For

    def __init__(self, node: For, nsp: Namespace, nsp_global: NamespaceGlobal):
        super().__init__(node, nsp, nsp_global)
        self.node = node

        self.converted_body = []
        self.converted_orelse = []

        self.break_cnt = 0
        self.interrupt_cnt = 0

        # flow-control vars
        self.flow_ctrl_wrapped_iter_expr = Name(id=ol_name(OL_WRAPPED_ITER))
        self.flow_ctrl_interrupt_expr = Name(id=ol_name(OL_INTERRUPT))
        self.flow_ctrl_interrupt_used = False
        self.interrupt_node_bodies = []

        self.nsp.loop_stack.append(self)

    def get_result(self) -> list[AST]:
        # if no break/continue/return used
        # use the simplest list comprehension
        if self.interrupt_cnt == 0 and len(self.node.orelse) == 0:
            return [
                ListComp(
                    elt=utils.list_wrapper(self.converted_body),
                    generators=[
                        comprehension(
                            target=self.node.target,
                            iter=expr_transf(self.nsp, self.node.iter),
                            ifs=[],
                            is_async=0,
                        )
                    ],
                )
            ]

        for_loop_final = []

        # init the flow-control vars
        if self.flow_ctrl_interrupt_used:
            self.converted_body.insert(
                0,
                NamedExpr(
                    target=self.flow_ctrl_interrupt_expr,
                    value=Constant(value=False),
                ),
            )
            # inject flow_ctrl_interrupt_expr to interrupt nodes
            for interrupt_body in self.interrupt_node_bodies:
                interrupt_body.append(
                    NamedExpr(
                        target=self.flow_ctrl_interrupt_expr,
                        value=Constant(value=True),
                    )
                )

        # we don't need use iter_wrapper
        # if we don't use break
        if self.break_cnt == 0:
            for_loop_iter = expr_transf(self.nsp, self.node.iter)
        else:
            from .presets import iter_wrapper_name

            self.nsp_global.use_preset_iter_wrapper = True
            for_loop_iter = self.flow_ctrl_wrapped_iter_expr
            iter_wrapper_instance = NamedExpr(
                target=self.flow_ctrl_wrapped_iter_expr,
                value=Call(
                    func=iter_wrapper_name,
                    args=[expr_transf(self.nsp, self.node.iter)],
                    keywords=[],
                ),
            )
            for_loop_final.append(iter_wrapper_instance)

        # "orelse" runs if there's no break
        if self.break_cnt:
            for_loop_orelse = IfExp(
                test=UnaryOp(
                    op=Not(),
                    operand=Attribute(
                        value=self.flow_ctrl_wrapped_iter_expr,
                        attr="_break",
                        ctx=Load(),
                    ),
                ),
                body=utils.list_wrapper(self.converted_orelse),
                orelse=Constant(value=...),
            )
        else:
            for_loop_orelse = utils.list_wrapper(self.converted_orelse)

        # the main body of the oneliner for loop
        for_loop_body = ListComp(
            elt=utils.list_wrapper(self.converted_body),
            generators=[
                comprehension(
                    target=self.node.target,
                    iter=for_loop_iter,
                    ifs=[],
                    is_async=0,
                )
            ],
        )

        # assemble all parts together and return

        for_loop_final.append(for_loop_body)
        # if orelse is not empty, add orelse
        if self.converted_orelse:
            for_loop_final.append(for_loop_orelse)

        return for_loop_final


class PendingBreak(PendingNode):
    def __init__(self, node: Break, nsp: Namespace, nsp_global: NamespaceGlobal):
        super().__init__(node, nsp, nsp_global)
        if len(self.nsp.loop_stack) == 0:
            raise SyntaxError(
                utils.ast_debug_info(node) + "'break' is not inside a loop"
            )

        self.loop = self.nsp.loop_stack[-1]
        self.loop.break_cnt += 1
        self.loop.interrupt_cnt += 1

    def get_result(self) -> list[AST]:
        return_value: list[AST] = []
        if isinstance(self.loop, PendingWhile):
            return_value.append(
                NamedExpr(
                    target=self.loop.flow_ctrl_break_expr,
                    value=Constant(value=True),
                )
            )
        elif isinstance(self.loop, PendingFor):
            return_value.append(
                Call(
                    func=Name(id="setattr", ctx=Load()),
                    args=[
                        self.loop.flow_ctrl_wrapped_iter_expr,
                        Constant(value="_break"),
                        Constant(value=True),
                    ],
                    keywords=[],
                )
            )
        self.loop.interrupt_node_bodies.append(return_value)
        return [List(elts=return_value, ctx=Load())]


class PeindingContinue(PendingNode):
    def __init__(self, node: Continue, nsp: Namespace, nsp_global: NamespaceGlobal):
        super().__init__(node, nsp, nsp_global)
        if len(self.nsp.loop_stack) == 0:
            raise SyntaxError(
                utils.ast_debug_info(node) + "'continue' is not inside a loop"
            )

        self.loop = self.nsp.loop_stack[-1]
        self.loop.interrupt_cnt += 1

    def get_result(self) -> list[AST]:
        return_value = []
        self.loop.interrupt_node_bodies.append(return_value)
        return [List(elts=return_value, ctx=Load())]


class PendingPass(PendingNode):
    def get_result(self) -> list[AST]:
        return [Constant(value=...)]


class PendingAssign(PendingNode):
    def __init__(
        self, node: Assign | AnnAssign, nsp: Namespace, nsp_global: NamespaceGlobal
    ):
        super().__init__(node, nsp, nsp_global)
        self.node = node

    def assign_auto(self, target: AST, value: expr) -> list[AST]:
        if isinstance(target, Name):
            return [self.assign_name(target, value)]
        elif isinstance(target, Attribute):
            return [self.assign_attribute(target, value)]
        elif isinstance(target, Subscript):
            return [self.assign_subscript(target, value)]
        elif isinstance(target, (Tuple, List)):
            return self.assign_tuple_list(target, value)
        else:
            raise NotImplementedError(f"Unknown assignment target: {type(target)}")

    def assign_name(self, target: Name, value: expr):
        return self.nsp.get_assign(target.id, value)

    def assign_subscript(self, target: Subscript, value: expr):
        _slice = target.slice
        if isinstance(_slice, Slice):
            _slice = utils.convert_slice(_slice)

        return Call(
            func=Attribute(
                value=expr_transf(self.nsp, target.value),
                attr="__setitem__",
                ctx=Load(),
            ),
            args=[_slice, value],
            keywords=[],
        )

    def assign_attribute(self, target: Attribute, value: expr) -> AST:
        return Call(
            func=Name(id="setattr", ctx=Load()),
            args=[
                expr_transf(self.nsp, target.value),
                Constant(value=target.attr),
                value,
            ],
            keywords=[],
        )

    def assign_tuple_list(self, target: Tuple | List, value: expr) -> list[AST]:
        """
        Recursion warning
        """
        return_list = []
        have_starred = False
        for index, sub_target in enumerate(target.elts):
            if isinstance(sub_target, Starred):
                if have_starred:
                    raise SyntaxError(
                        utils.ast_debug_info(sub_target)
                        + "multiple starred expressions in assignment"
                    )
                have_starred = True

                sub_target = sub_target.value

                slice_upper = Constant(value=index - len(target.elts) + 1)
                if slice_upper.value == 0:
                    slice_upper = None

                value_subscript = Call(
                    func=Name(id="list", ctx=Load()),
                    args=[
                        Subscript(
                            value=value,
                            slice=Slice(
                                lower=Constant(value=index),
                                upper=slice_upper,
                            ),
                            ctx=Load(),
                        )
                    ],
                    keywords=[],
                )
            else:
                if not have_starred:
                    _slice = Constant(value=index)
                else:
                    _slice = Constant(value=index - len(target.elts))

                value_subscript = Subscript(
                    value=value,
                    slice=_slice,
                    ctx=Load(),
                )

            return_list.extend(self.assign_auto(sub_target, value_subscript))
        return return_list

    def get_result(self) -> list[AST]:
        if self.node.value is None:
            return []

        return_list = []
        assign_value = expr_transf(self.nsp, self.node.value)

        if isinstance(self.node, AnnAssign):
            assign_targets = [self.node.target]
        else:
            assign_targets = self.node.targets

        if len(assign_targets) == 1 and not isinstance(
            assign_targets[0], (Tuple, List)
        ):
            # if there is only one target (not tuple or list)
            # wo don't need tmp var
            return_value = self.assign_auto(assign_targets[0], assign_value)
            assert len(return_value) == 1
            return return_value

        # save the assign value to a tmp var
        # to make sure the value expr only runs once.
        tmp_value_name = Name(id=ol_name(OL_ASSIGN_TMP))
        return_list.append(
            NamedExpr(target=tmp_value_name, value=assign_value),
        )
        for target in assign_targets:
            return_list.extend(self.assign_auto(target, tmp_value_name))

        return return_list


class PendingAugAssign(PendingNode):
    def __init__(self, node: AugAssign, nsp: Namespace, nsp_global: NamespaceGlobal):
        super().__init__(node, nsp, nsp_global)
        self.node = node

    _op_dict: dict[type[operator], str] = {
        Add: "__iadd__",
        BitAnd: "__iand__",
        FloorDiv: "__ifloordiv__",
        LShift: "__ilshift__",
        Mod: "__imod__",
        Mult: "__imul__",
        MatMult: "__imatmul__",
        BitOr: "__ior__",
        Pow: "__ipow__",
        RShift: "__irshift__",
        Sub: "__isub__",
        Div: "__itruediv__",
        BitXor: "__ixor__",
    }

    def _aug_assign_expr(
        self, target: expr, op: operator, value: AST, fallback: AST | None = None
    ):
        op_name = self._op_dict[type(op)]
        if fallback is None:
            fallback = NamedExpr(
                target=target, value=BinOp(left=target, op=op, right=value)
            )
        return IfExp(
            test=Call(
                func=Name(id="hasattr", ctx=Load()),
                args=[target, Constant(value=op_name)],
                keywords=[],
            ),
            body=Call(
                func=Attribute(value=target, attr=op_name, ctx=Load()),
                args=[value],
                keywords=[],
            ),
            orelse=fallback,
        )

    def get_result(self) -> list[AST]:
        return_list = []
        tmp_target_name = Name(id=ol_name(OL_AUGASSIGN_TMP))
        assign_value = expr_transf(self.nsp, self.node.value)
        if isinstance(self.node.target, Name):
            target = self.nsp.get_load_name(self.node.target.id)
            return [
                self._aug_assign_expr(
                    target,
                    self.node.op,
                    assign_value,
                    fallback=self.nsp.get_assign(
                        self.node.target.id,
                        BinOp(left=target, op=self.node.op, right=assign_value),
                    ),
                )
            ]
        elif isinstance(self.node.target, Subscript):
            # todo: could be optimized if slice is const
            tmp_slice_name = Name(id=ol_name(OL_AUGASSIGN_SLICE_TMP))
            target = self.node.target
            subscript_parent = expr_transf(self.nsp, target.value)

            slice_expr = target.slice
            if isinstance(slice_expr, Slice):
                slice_expr = utils.convert_slice(slice_expr)

            # save slice expr to a tmp
            return_list.append(
                NamedExpr(
                    target=tmp_slice_name,
                    value=expr_transf(self.nsp, slice_expr),
                )
            )

            # load subscript value to a tmp
            return_list.append(
                NamedExpr(
                    target=tmp_target_name,
                    value=Subscript(
                        value=subscript_parent,
                        slice=tmp_slice_name,
                        ctx=Load(),
                    ),
                )
            )

            # aug assign to tmp
            _assign_body = self._aug_assign_expr(
                tmp_target_name, self.node.op, assign_value
            )

            # assign the tmp back to original subscript
            return_list.append(
                Call(
                    func=Attribute(
                        value=subscript_parent,
                        attr="__setitem__",
                        ctx=Load(),
                    ),
                    args=[tmp_slice_name, _assign_body],
                    keywords=[],
                )
            )
        elif isinstance(self.node.target, Attribute):
            target = self.node.target
            attr_parent = expr_transf(self.nsp, target.value)
            return_list.append(
                NamedExpr(
                    target=tmp_target_name,
                    value=Attribute(
                        value=attr_parent,
                        attr=target.attr,
                        ctx=Load(),
                    ),
                )
            )
            _assign_body = self._aug_assign_expr(
                tmp_target_name, self.node.op, assign_value
            )
            return_list.append(
                Call(
                    func=Name(id="setattr", ctx=Load()),
                    args=[
                        attr_parent,
                        Constant(value=target.attr),
                        _assign_body,
                    ],
                    keywords=[],
                )
            )
        else:
            raise NotImplementedError(
                f"Unknown augmented assignment target: {type(self.node.target)}"
            )
        return return_list


class PendingFunctionDef(_PendingCompoundStmt):
    has_internal_namespace = True
    internal_nsp: NamespaceFunction

    def __init__(self, node: FunctionDef, nsp: Namespace, nsp_global: NamespaceGlobal):
        super().__init__(node, nsp, nsp_global)
        self.node = node

        for tmp_nsp in self.nsp.inner_nsp:
            if (
                tmp_nsp.symt.get_lineno() == node.lineno
                and tmp_nsp.symt.get_name() == node.name
            ):
                assert isinstance(tmp_nsp, NamespaceFunction)
                self.internal_nsp = tmp_nsp
                break
        else:
            raise RuntimeError("Namespace not found")

        # copy args and filter annotations
        original_args = node.args
        self.converted_args = converted_args = arguments()

        converted_args.posonlyargs = []
        for _arg in original_args.posonlyargs:
            converted_args.posonlyargs.append(arg(arg=_arg.arg))
        converted_args.args = []
        for _arg in original_args.args:
            converted_args.args.append(arg(arg=_arg.arg))
        converted_args.kwonlyargs = []
        for _arg in original_args.kwonlyargs:
            converted_args.kwonlyargs.append(arg(arg=_arg.arg))
        if original_args.vararg is not None:
            converted_args.vararg = arg(arg=original_args.vararg.arg)
        if original_args.kwarg is not None:
            converted_args.kwarg = arg(arg=original_args.kwarg.arg)

        converted_args.defaults = []
        for default_expr in original_args.defaults:
            converted_args.defaults.append(expr_transf(self.nsp, default_expr))

        converted_args.kw_defaults = []
        for kw_default_expr in original_args.kw_defaults:
            if kw_default_expr is None:
                converted_args.kw_defaults.append(None)
                continue
            converted_args.kw_defaults.append(expr_transf(self.nsp, kw_default_expr))

        self.converted_body = []

    def get_internal_namespace(self):
        return self.internal_nsp

    def _iter_nodes(self) -> typing.Generator[AST, list[AST], None]:
        yield from self._iter_branch(
            self.converted_body,
            self.node.body,
            lambda: self.internal_nsp.return_cnt,
            self.internal_nsp.get_flow_ctrl_expr,
        )

    def get_result(self) -> list[AST]:
        body: list[AST] = []
        body.append(
            NamedExpr(
                target=self.internal_nsp.return_value_expr,
                value=Constant(value=None),
            )
        )

        if self.internal_nsp.zero_arg_super_used:
            # inject free __class__
            body.append(Name(id="__class__", ctx=Load()))

        if self.internal_nsp.flow_ctrl_return_used:
            body.append(
                NamedExpr(
                    target=self.internal_nsp.flow_ctrl_return_expr,
                    value=Constant(value=False),
                )
            )
            # inject flow_ctrl_return to interrupt nodes
            for return_node_body in self.internal_nsp.return_node_bodies:
                return_node_body.append(
                    NamedExpr(
                        target=self.internal_nsp.flow_ctrl_return_expr,
                        value=Constant(value=True),
                    )
                )

        if len(self.internal_nsp.inner_nonlocal_names):
            nonlocal_dict_keys = []
            nonlocal_dict_values = []
            for nonlocal_param in self.internal_nsp.nonlocal_parameters:
                nonlocal_dict_keys.append(Constant(value=nonlocal_param))
                nonlocal_dict_values.append(Name(id=nonlocal_param, ctx=Load()))
            body.append(
                NamedExpr(
                    target=self.internal_nsp.nonlocal_dict_expr,
                    value=Dict(keys=nonlocal_dict_keys, values=nonlocal_dict_values),
                )
            )

        body.extend(self.converted_body)
        body.append(self.internal_nsp.return_value_expr)

        body_expr = Lambda(
            args=self.converted_args,
            body=Subscript(
                value=utils.list_wrapper(body),
                slice=Constant(value=-1),
                ctx=Load(),
            ),
        )
        for dec_expr in reversed(self.node.decorator_list):
            body_expr = Call(
                func=expr_transf(self.nsp, dec_expr),
                args=[body_expr],
                keywords=[],
            )

        if self.internal_nsp.is_method and self.node.name == "__init_subclass__":
            # We need to add a @classmethod for __init_subclass__
            # that's really weird, but really solves problem
            body_expr = Call(
                func=Name(id="classmethod", ctx=Load()),
                args=[body_expr],
                keywords=[],
            )

        return [self.nsp.get_assign(self.node.name, body_expr)]


class PendingReturn(PendingNode):
    def __init__(self, node: Return, nsp: Namespace, nsp_global: NamespaceGlobal):
        if not isinstance(nsp, NamespaceFunction):
            raise SyntaxError(utils.ast_debug_info(node) + "'return' outside function")

        super().__init__(node, nsp, nsp_global)
        self.node = node
        self.nsp: NamespaceFunction

        self.nsp.return_cnt += 1
        for loop in self.nsp.loop_stack:
            loop.break_cnt += 1
            loop.interrupt_cnt += 1

    def get_result(self) -> list[AST]:
        return_list: list[AST] = []
        if self.node.value is not None:
            return_list.append(
                NamedExpr(
                    target=self.nsp.return_value_expr,
                    value=expr_transf(self.nsp, self.node.value),
                )
            )

        for loop in self.nsp.loop_stack:
            if isinstance(loop, PendingWhile):
                return_list.append(
                    NamedExpr(
                        target=loop.flow_ctrl_break_expr,
                        value=Constant(value=True),
                    )
                )
            elif isinstance(loop, PendingFor):
                return_list.append(
                    Call(
                        func=Name(id="setattr", ctx=Load()),
                        args=[
                            loop.flow_ctrl_wrapped_iter_expr,
                            Constant(value="_break"),
                            Constant(value=True),
                        ],
                        keywords=[],
                    )
                )
        for loop in self.nsp.loop_stack:
            loop.interrupt_node_bodies.append(return_list)
        self.nsp.return_node_bodies.append(return_list)
        return [List(elts=return_list, ctx=Load())]


class PendingGlobal(PendingNode):
    def get_result(self) -> list[AST]:
        return []


class PendingNonlocal(PendingNode):
    def get_result(self) -> list[AST]:
        return []


class PendingClassDef(_PendingCompoundStmt):
    has_internal_namespace = True
    internal_nsp: NamespaceClass

    def __init__(self, node: ClassDef, nsp: Namespace, nsp_global: NamespaceGlobal):
        super().__init__(node, nsp, nsp_global)
        self.node = node

        for tmp_nsp in self.nsp.inner_nsp:
            if (
                tmp_nsp.symt.get_lineno() == node.lineno
                and tmp_nsp.symt.get_name() == node.name
            ):
                assert isinstance(tmp_nsp, NamespaceClass)
                self.internal_nsp = tmp_nsp
                break
        else:
            raise RuntimeError("Namespace not found")

        self.converted_body = []

    def get_internal_namespace(self) -> Namespace:
        return self.internal_nsp

    def _iter_nodes(self) -> typing.Generator[AST, list[AST], None]:
        yield from self._iter_branch(
            self.converted_body,
            self.node.body,
            lambda: 0,  # class doesn't have any flow control
            utils.never_call,
        )

    def get_result(self) -> list[AST]:
        return_list: list[AST] = []

        class_bases = [expr_transf(self.nsp, _expr) for _expr in self.node.bases]

        metaclass_expr = None
        class_keywords = []
        for _keyword in self.node.keywords:
            if _keyword.arg == "metaclass":
                # filter the metaclass keyword
                metaclass_expr = expr_transf(self.nsp, _keyword.value)
                continue
            class_keywords.append(
                keyword(
                    arg=_keyword.arg,
                    value=expr_transf(self.nsp, _keyword.value),
                )
            )

        if metaclass_expr is None:
            metaclass_expr = Name(id="type", ctx=Load())

        return_list.append(
            self.nsp.get_assign(
                self.node.name,
                Call(
                    func=metaclass_expr,
                    args=[
                        Constant(value=self.node.name),
                        Tuple(elts=class_bases, ctx=Load()),
                        Dict(keys=[], values=[]),
                    ],
                    keywords=class_keywords,
                ),
            )
        )

        class_body: list[AST] = []
        class_body.append(
            NamedExpr(  # one step of injecting the __class__ cell
                target=Name(id="__class__", ctx=Store()),
                value=self.nsp.get_load_name(self.node.name),
            )
        )
        class_body.append(
            NamedExpr(
                target=self.internal_nsp.class_member_dict_expr,
                value=Dict(keys=[], values=[]),
            )
        )
        class_body.extend(self.converted_body)
        class_body.append(self.internal_nsp.class_member_dict_expr)

        loader_name = ol_name(OL_CLASS_LOADER)

        return_list.append(
            NamedExpr(
                target=Name(id=loader_name, ctx=Store()),
                value=Lambda(
                    args=arguments(
                        posonlyargs=[],
                        args=[],
                        kwonlyargs=[],
                        kw_defaults=[],
                        defaults=[],
                    ),
                    body=Subscript(
                        value=List(elts=class_body, ctx=Load()),
                        slice=UnaryOp(op=USub(), operand=Constant(value=1)),
                        ctx=Load(),
                    ),
                ),
            )
        )

        load_class = ListComp(
            elt=Call(
                func=Name(id="setattr", ctx=Load()),
                args=[
                    self.nsp.get_load_name(self.node.name),
                    Name(id="k", ctx=Load()),
                    Name(id="v", ctx=Load()),
                ],
                keywords=[],
            ),
            generators=[
                comprehension(
                    target=Tuple(
                        elts=[Name(id="k", ctx=Store()), Name(id="v", ctx=Store())],
                        ctx=Store(),
                    ),
                    iter=Call(
                        func=Attribute(
                            value=Call(
                                func=Name(id=loader_name, ctx=Load()),
                                args=[],
                                keywords=[],
                            ),
                            attr="items",
                            ctx=Load(),
                        ),
                        args=[],
                        keywords=[],
                    ),
                    ifs=[],
                    is_async=0,
                )
            ],
        )
        return_list.append(load_class)
        return return_list


class PendingImport(PendingNode):
    def __init__(self, node: Import, nsp: Namespace, nsp_global: NamespaceGlobal):
        super().__init__(node, nsp, nsp_global)
        self.node = node
        self.nsp_global.use_importlib = True

    def get_result(self) -> list[AST]:
        result = []
        for _alias in self.node.names:
            if _alias.asname is None:
                asname = _alias.name
            else:
                asname = _alias.asname

            result.append(
                self.nsp.get_assign(
                    asname,
                    Call(
                        func=Attribute(
                            value=Name(id="importlib", ctx=Load()),
                            attr="import_module",
                        ),
                        args=[Constant(value=_alias.name)],
                        keywords=[],
                    ),
                )
            )

        return result


class PendingImportFrom(PendingNode):
    def __init__(self, node: ImportFrom, nsp: Namespace, nsp_global: NamespaceGlobal):
        super().__init__(node, nsp, nsp_global)
        self.node = node

    def get_result(self) -> list[AST]:
        result = []
        tmp_mod_name = ol_name(OL_IMPORT_TMP)

        if self.node.module is None:
            mod_name = ""
        else:
            mod_name = self.node.module

        from_list = []
        for _alias in self.node.names:
            from_list.append(Constant(value=_alias.name))

        import_body = NamedExpr(
            target=Name(id=tmp_mod_name, ctx=Store()),
            value=Call(
                func=Name(id="__import__", ctx=Load()),
                args=[
                    Constant(value=mod_name),
                    Call(func=Name(id="globals", ctx=Load()), args=[], keywords=[]),
                    Call(func=Name(id="locals", ctx=Load()), args=[], keywords=[]),
                    List(elts=from_list, ctx=Load()),
                    Constant(value=self.node.level),
                ],
                keywords=[],
            ),
        )
        result.append(import_body)

        for _alias in self.node.names:
            if _alias.asname is None:
                asname = _alias.name
            else:
                asname = _alias.asname

            if _alias.name == "*":
                raise RuntimeError(
                    "Unable to convert 'from ... import *'"
                )  # pragma: no cover

            result.append(
                self.nsp.get_assign(
                    asname,
                    Attribute(
                        value=Name(id=tmp_mod_name, ctx=Store()),
                        attr=_alias.name,
                    ),
                )
            )

        return result



from typing import TypeAlias

import oneliner.utils as utils

_ol_reserved_name: TypeAlias = str


OL_BREAK: _ol_reserved_name = "__ol_break_{}"
OL_INTERRUPT: _ol_reserved_name = "__ol_interrupt_{}"
OL_WRAPPED_ITER: _ol_reserved_name = "__ol_it_{}"
OL_ITER_WRAPPER: _ol_reserved_name = "__ol_iter_wrapper"  # don't need format here
OL_ASSIGN_TMP: _ol_reserved_name = "__ol_assign_{}"
OL_AUGASSIGN_TMP: _ol_reserved_name = "__ol_augass_{}"
OL_AUGASSIGN_SLICE_TMP: _ol_reserved_name = "__ol_sllice_{}"
OL_RETURN_VALUE: _ol_reserved_name = "__ol_retv_{}"
OL_RETURN: _ol_reserved_name = "__ol_ret_{}"
OL_NONLOCAL_DICT: _ol_reserved_name = "__ol_nonlocal_{}"
OL_CLASS_DICT: _ol_reserved_name = "__ol_classnsp_{}"
OL_CLASS_LOADER: _ol_reserved_name = "__ol_loader_{}"
OL_IMPORT_TMP: _ol_reserved_name = "__ol_mod_{}"


def ol_name(name: _ol_reserved_name):
    return name.format(utils.unique_id())



"""
Preset "iter_wrapper"
Used in for loop to handle "break"s
"""

# This is the original code of __ol_iter_wrapper
"""
class __ol_iter_wrapper:
    def __init__(self,it):
        self.it = iter(it)
        self._break = False
    
    def __iter__(self):
        return self
    
    def __next__(self):
        if self._break:
            next(iter([]))
        else:
            return next(self.it)
"""

# This is the oneliner version __ol_iter_wrapper
"""
(
    __ol_iter_wrapper := type(
        "__ol_iter_wrapper",
        (),
        {
            "__init__": (
                lambda self, it: [
                    setattr(self, "it", iter(it)),
                    setattr(self, "_break", False),
                    None,
                ][-1]
            ),
            "__iter__": (lambda self: self),
            "__next__": (lambda self: next(iter([])) if self._break else next(self.it)),
        },
    )
)
"""

from ast import *

from oneliner.reserved_identifiers import OL_ITER_WRAPPER

iter_wrapper_name = Name(id=OL_ITER_WRAPPER, ctx=Load())
iter_wrapper_body = NamedExpr(
    target=Name(id=OL_ITER_WRAPPER, ctx=Store()),
    value=Call(
        func=Name(id="type", ctx=Load()),
        args=[
            Constant(value=OL_ITER_WRAPPER),
            Tuple(elts=[], ctx=Load()),
            Dict(
                keys=[
                    Constant(value="__init__"),
                    Constant(value="__iter__"),
                    Constant(value="__next__"),
                ],
                values=[
                    Lambda(
                        args=arguments(
                            posonlyargs=[],
                            args=[arg(arg="self"), arg(arg="it")],
                            kwonlyargs=[],
                            kw_defaults=[],
                            defaults=[],
                        ),
                        body=Subscript(
                            value=List(
                                elts=[
                                    Call(
                                        func=Name(id="setattr", ctx=Load()),
                                        args=[
                                            Name(id="self", ctx=Load()),
                                            Constant(value="it"),
                                            Call(
                                                func=Name(id="iter", ctx=Load()),
                                                args=[Name(id="it", ctx=Load())],
                                                keywords=[],
                                            ),
                                        ],
                                        keywords=[],
                                    ),
                                    Call(
                                        func=Name(id="setattr", ctx=Load()),
                                        args=[
                                            Name(id="self", ctx=Load()),
                                            Constant(value="_break"),
                                            Constant(value=False),
                                        ],
                                        keywords=[],
                                    ),
                                    Constant(value=None),
                                ],
                                ctx=Load(),
                            ),
                            slice=UnaryOp(op=USub(), operand=Constant(value=1)),
                            ctx=Load(),
                        ),
                    ),
                    Lambda(
                        args=arguments(
                            posonlyargs=[],
                            args=[arg(arg="self")],
                            kwonlyargs=[],
                            kw_defaults=[],
                            defaults=[],
                        ),
                        body=Name(id="self", ctx=Load()),
                    ),
                    Lambda(
                        args=arguments(
                            posonlyargs=[],
                            args=[arg(arg="self")],
                            kwonlyargs=[],
                            kw_defaults=[],
                            defaults=[],
                        ),
                        body=IfExp(
                            test=Attribute(
                                value=Name(id="self", ctx=Load()),
                                attr="_break",
                                ctx=Load(),
                            ),
                            body=Call(
                                func=Name(id="next", ctx=Load()),
                                args=[
                                    Call(
                                        func=Name(id="iter", ctx=Load()),
                                        args=[List(elts=[], ctx=Load())],
                                        keywords=[],
                                    )
                                ],
                                keywords=[],
                            ),
                            orelse=Call(
                                func=Name(id="next", ctx=Load()),
                                args=[
                                    Attribute(
                                        value=Name(id="self", ctx=Load()),
                                        attr="it",
                                        ctx=Load(),
                                    )
                                ],
                                keywords=[],
                            ),
                        ),
                    ),
                ],
            ),
        ],
        keywords=[],
    ),
)



import itertools
import symtable
import sys
from ast import *

from oneliner.reserved_identifiers import *

__all__ = [
    "generate_nsp",
    "Namespace",
    "NamespaceGlobal",
    "NamespaceFunction",
    "NamespaceClass",
]


class Namespace:
    symt: symtable.SymbolTable
    outer_nsp: "Namespace"
    inner_nsp: list["Namespace"]

    loop_stack: list["oneliner.pending_nodes._PendingLoop"]
    comp_stack: list["oneliner.expr_transform.PendingComp"]

    @classmethod
    def _generate(cls):
        raise NotImplementedError()  # pragma: no cover

    def __init__(self):
        self.loop_stack = []
        self.comp_stack = []
        self.inner_nsp = []

    def get_assign(self, name: str, value_expr: expr) -> expr:
        """
        In different namespaces,
        different types of names (global/local) have different assign methods.
        Use this to get the assign method for name except internal name
        """
        raise NotImplementedError()  # pragma: no cover

    def get_load_name(self, name: str) -> expr:
        """
        nonlocal names have different method to be load.
        Use this to get the mtehod to load the name
        """
        raise NotImplementedError()  # pragma: no cover


class NamespaceGlobal(Namespace):
    use_itertools: bool
    use_importlib: bool
    use_preset_iter_wrapper: bool

    @classmethod
    def _generate(cls, symt: symtable.SymbolTable):
        self = cls()
        self.symt = symt
        return self

    def __init__(self):
        super().__init__()

        self.use_itertools = False
        self.use_importlib = False
        self.use_preset_iter_wrapper = False

    def get_assign(self, name: str, value_expr: expr) -> NamedExpr:
        return NamedExpr(target=Name(id=name, ctx=Store()), value=value_expr)

    def get_load_name(self, name: str) -> Name:
        return Name(id=name, ctx=Load())


class NamespaceFunction(Namespace):
    inner_nonlocal_names: set[str]  # names that is nonlocal in INNER namespace
    nonlocal_parameters: set[str]  # parameters that is nonlocal in INNER namespace
    outer_nonlocal_map: dict[str, "NamespaceFunction"]
    # keys   --> nonlocal names of THIS namespace
    # values --> where the nonlocal name was born

    is_method: bool  # whether the function is a method
    zero_arg_super_used: bool  # whether the method uses a zero-argument super

    symt: symtable.Function  # symbol table of this namespace

    @classmethod
    def _generate(cls, symt: symtable.Function, stack: list[Namespace]):
        # don't push/pop the stack in this function

        self = cls()
        self.symt = symt

        self.outer_nsp = stack[-1]
        self.outer_nsp.inner_nsp.append(self)
        self.inner_nonlocal_names = set()
        self.nonlocal_parameters = set()
        self.outer_nonlocal_map = {}
        self.is_method = False
        self.zero_arg_super_used = False

        if (
            isinstance(stack[-1], NamespaceClass)
            and self.symt.get_name() in stack[-1].symt.get_methods()
        ):
            self.is_method = True

        for nonlocal_free in itertools.chain(
            # treat 'free's as 'nonlocal's
            self.symt.get_frees(),
            self.symt.get_nonlocals(),
        ):
            if self.is_method and nonlocal_free == "__class__":
                # methods may have implicit reference the __class__ (PEP-3135)
                # which is not need here
                self.zero_arg_super_used = True
                break

            for outer in reversed(stack):
                if isinstance(outer, NamespaceClass):
                    continue
                assert isinstance(outer, NamespaceFunction)

                # free/nonlocal inevitablely exist in outer function namespace
                # so check is not need here.
                outer_symbol = outer.symt.lookup(nonlocal_free)
                if (
                    outer_symbol.is_assigned()
                    or outer_symbol.is_parameter()
                    and not outer_symbol.is_global()
                ):
                    outer.inner_nonlocal_names.add(nonlocal_free)
                    self.outer_nonlocal_map[nonlocal_free] = outer
                    if outer_symbol.is_parameter():
                        outer.nonlocal_parameters.add(nonlocal_free)
                    break
            else:
                raise RuntimeError(  # pragma: no cover
                    f"Unable to search the origin of nonlocal/free '{nonlocal_free}'"
                )

        return self

    def __init__(self):
        super().__init__()

        self.return_cnt = 0

        self.return_value_expr = Name(id=ol_name(OL_RETURN_VALUE))
        self.flow_ctrl_return_expr = Name(id=ol_name(OL_RETURN))
        self.flow_ctrl_return_used = False
        self.return_node_bodies: list[list[AST]] = (
            []
        )  # list of bodies of converted return nodes

        # use a dict to emulate the behavior of nonlocal
        self.nonlocal_dict_expr = Name(id=ol_name(OL_NONLOCAL_DICT))

    def get_flow_ctrl_expr(self):
        self.flow_ctrl_return_used = True
        return self.flow_ctrl_return_expr

    def get_assign(self, name: str, value_expr: expr) -> expr:
        symbol = self.symt.lookup(name)
        if symbol.is_declared_global():
            return Call(
                func=Attribute(
                    value=Call(
                        func=Name(id="globals", ctx=Load()), args=[], keywords=[]
                    ),
                    attr="__setitem__",
                    ctx=Load(),
                ),
                args=[Constant(value=name), value_expr],
                keywords=[],
            )
        elif name in self.outer_nonlocal_map:
            outer = self.outer_nonlocal_map[name]
            return Call(
                func=Attribute(
                    value=outer.nonlocal_dict_expr, attr="__setitem__", ctx=Load()
                ),
                args=[Constant(value=name), value_expr],
                keywords=[],
            )
        elif name in self.inner_nonlocal_names:
            return Call(
                func=Attribute(
                    value=self.nonlocal_dict_expr, attr="__setitem__", ctx=Load()
                ),
                args=[Constant(value=name), value_expr],
                keywords=[],
            )
        else:
            return NamedExpr(
                target=Name(id=name, ctx=Store()),
                value=value_expr,
            )

    def get_load_name(self, name: str) -> expr:
        for comp in self.comp_stack:
            if name in comp.target_names:
                return Name(id=name, ctx=Load())

        if name in self.inner_nonlocal_names:
            return Subscript(
                value=self.nonlocal_dict_expr,
                slice=Constant(value=name),
                ctx=Load(),
            )
        elif name in self.outer_nonlocal_map:
            outer = self.outer_nonlocal_map[name]
            return Subscript(
                value=outer.nonlocal_dict_expr,
                slice=Constant(value=name),
                ctx=Load(),
            )
        else:  # globals or locals except free
            return Name(id=name, ctx=Load())


class NamespaceClass(Namespace):
    symt: symtable.Class

    # NamespaceClass doesn't have inner_nonlocal_names

    outer_nonlocal_map: dict[str, "NamespaceFunction"]
    # keys   --> nonlocal names of THIS namespace
    # values --> where the nonlocal name was born

    if sys.version_info < (3, 12):
        globals_used_in_comp: set[str]  # global names used in comprehensions

    @classmethod
    def _generate(cls, symt: symtable.Class, stack: list[Namespace]):
        # don't push/pop the stack in this function

        self = cls()
        self.symt = symt

        self.outer_nsp = stack[-1]
        self.outer_nsp.inner_nsp.append(self)
        self.outer_nonlocal_map = {}
        if sys.version_info < (3, 12):
            self.globals_used_in_comp = set()

        for symbol in self.symt.get_symbols():
            if not (symbol.is_nonlocal() or symbol.is_free()):
                continue
            nonlocal_free: str = symbol.get_name()
            # treat 'free's as 'nonlocal's
            for outer in reversed(stack):
                if isinstance(outer, NamespaceClass):
                    continue
                assert isinstance(outer, NamespaceFunction)

                # free/nonlocal inevitablely exist in outer function namespace
                # so check is not need here.
                outer_symbol = outer.symt.lookup(nonlocal_free)
                if (
                    outer_symbol.is_assigned()
                    or outer_symbol.is_parameter()
                    and not outer_symbol.is_global()
                ):
                    outer.inner_nonlocal_names.add(nonlocal_free)
                    self.outer_nonlocal_map[nonlocal_free] = outer
                    if outer_symbol.is_parameter():
                        outer.nonlocal_parameters.add(nonlocal_free)
                    break
            else:
                raise RuntimeError(  # pragma: no cover
                    f"Unable to search the origin of nonlocal/free '{nonlocal_free}'"
                )

        return self

    def __init__(self):
        super().__init__()
        self.class_member_dict_expr = Name(id=ol_name(OL_CLASS_DICT))

    def get_assign(self, name: str, value_expr: expr) -> expr:
        symbol = self.symt.lookup(name)
        if symbol.is_declared_global():
            return Call(
                func=Attribute(
                    value=Call(
                        func=Name(id="globals", ctx=Load()), args=[], keywords=[]
                    ),
                    attr="__setitem__",
                    ctx=Load(),
                ),
                args=[Constant(value=name), value_expr],
                keywords=[],
            )
        elif name in self.outer_nonlocal_map:
            outer = self.outer_nonlocal_map[name]
            return Call(
                func=Attribute(
                    value=outer.nonlocal_dict_expr, attr="__setitem__", ctx=Load()
                ),
                args=[Constant(value=name), value_expr],
                keywords=[],
            )
        else:
            # assign to a class member
            return Call(
                func=Attribute(
                    value=self.class_member_dict_expr, attr="__setitem__", ctx=Load()
                ),
                args=[Constant(value=name), value_expr],
                keywords=[],
            )

    def get_load_name(self, name: str) -> expr:
        for comp in self.comp_stack:
            if name in comp.target_names:
                return Name(id=name, ctx=Load())

        if sys.version_info < (3, 12) and name in self.globals_used_in_comp:
            return Name(id=name, ctx=Load())

        symbol = self.symt.lookup(name)
        if name in self.outer_nonlocal_map:
            outer = self.outer_nonlocal_map[name]
            return Subscript(
                value=outer.nonlocal_dict_expr,
                slice=Constant(value=name),
                ctx=Load(),
            )
        elif symbol.is_global():
            return Name(id=name, ctx=Load())
        else:
            # a class member
            return Subscript(
                value=self.class_member_dict_expr,
                slice=Constant(value=name),
                ctx=Load(),
            )


if sys.version_info < (3, 12):

    def _comp_check(symt: symtable.Function):
        if symt.get_name() not in ["listcomp", "genexpr", "setcomp", "dictcomp"]:
            return False
        if ".0" not in symt.get_parameters():
            return False
        return True


def update_globals_from_lambda_or_comp(symt: symtable.Function, stack: list[Namespace]):
    if not isinstance(stack[-1], NamespaceClass):
        return

    _globals: set[str] = set()
    comp_stack = [symt]
    while comp_stack:
        symt = comp_stack.pop()
        for symbol in symt.get_symbols():
            if symbol.is_global():
                _globals.add(symbol.get_name())
        for child_symt in symt.get_children():
            assert isinstance(child_symt, symtable.Function)
            comp_stack.append(child_symt)
    stack[-1].globals_used_in_comp.update(_globals)


def generate_nsp(symt: symtable.SymbolTable):
    walk_stack = []
    root = NamespaceGlobal._generate(symt)
    generate_stack: list[Namespace] = [root]

    walk_stack.append(iter(symt.get_children()))
    while walk_stack:
        try:
            child_symt = next(walk_stack[-1])
        except StopIteration:
            walk_stack.pop()
            generate_stack.pop()
        else:
            if isinstance(child_symt, symtable.Function):
                if child_symt.get_name() == "lambda":
                    update_globals_from_lambda_or_comp(child_symt, generate_stack)
                    continue
                if sys.version_info < (3, 12):
                    if _comp_check(child_symt):
                        update_globals_from_lambda_or_comp(child_symt, generate_stack)
                        continue

                generate_stack.append(
                    NamespaceFunction._generate(child_symt, generate_stack)
                )
            elif isinstance(child_symt, symtable.Class):
                generate_stack.append(
                    NamespaceClass._generate(child_symt, generate_stack)
                )
            else:  # pragma: no cover
                raise RuntimeError("Unknown type of child symbol table")
            walk_stack.append(iter(child_symt.get_children()))
    return root


# fix error caused by circular import
import oneliner.expr_transform
import oneliner.pending_nodes



import typing
from ast import *

from oneliner.namespaces import Namespace

__all__ = [
    "expr_transf",
]


class PendingExprGeneric:
    def __init__(self, node: expr):
        self.node = node
        self.converted_dict: dict[str, expr | list[expr]] = {}

        self.iter_fields = self._iter_fields()

    def _iter_fields(self):
        for field_name in self.node._fields:
            if not hasattr(self.node, field_name):
                continue
            field = getattr(self.node, field_name)
            if isinstance(field, expr):
                self.converted_dict[field_name] = yield field
            elif isinstance(field, list):
                converted_list = self.converted_dict[field_name] = []
                for _node in field:
                    converted_list.append((yield _node))
            else:
                self.converted_dict[field_name] = field

    def get_result(self) -> expr:
        return type(self.node)(**self.converted_dict)


class PendingNamedExpr(PendingExprGeneric):
    def __init__(self, node: NamedExpr, nsp: Namespace):
        self.node = node
        self.nsp = nsp

        self.value = None

        self.iter_fields = self._iter_fields()

    def _iter_fields(self):
        self.value = yield self.node.value

    def get_result(self) -> expr:
        assert self.value is not None
        result = self.nsp.get_assign(self.node.target.id, self.value)
        if not isinstance(result, NamedExpr):
            result = Subscript(
                value=List(
                    elts=[
                        result,
                        self.node.target,
                    ],
                    ctx=Load(),
                ),
                slice=Constant(value=-1),
                ctx=Load(),
            )

        return result


class PendingName(PendingExprGeneric):
    def __init__(self, node: Name, nsp: Namespace):
        self.node = node
        self.nsp = nsp

        self.iter_fields = self._iter_fields()

    def _iter_fields(self):
        return
        yield

    def get_result(self) -> expr:
        if isinstance(self.node.ctx, Store):
            return Name(id=self.node.id, ctx=Store())
        return self.nsp.get_load_name(self.node.id)


_CompNode: typing.TypeAlias = ListComp | SetComp | DictComp | GeneratorExp


class PendingComp(PendingExprGeneric):
    target_names: set[str]
    node: _CompNode

    def __init__(self, node: _CompNode, nsp: Namespace):
        super().__init__(node)
        self.nsp = nsp
        self.target_names = set()

        for comp in self.node.generators:
            self.get_comp_target_names(comp.target)

        self.nsp.comp_stack.append(self)

    def get_result(self) -> expr:
        assert self.nsp.comp_stack[-1] is self
        self.nsp.comp_stack.pop()

        return super().get_result()

    def get_comp_target_names(self, target):
        """
        Recursion warning
        """
        if isinstance(target, Name):
            self.target_names.add(target.id)
        elif isinstance(target, (Tuple, List)):
            for sub_target in target.elts:
                self.get_comp_target_names(sub_target)
        else:  # pragma: no cover
            raise RuntimeError("Unknown comprehension target")


class ExpressionTransformer:
    def __init__(self, nsp: Namespace):
        self.pending_stack: list[PendingExprGeneric] = []
        self.nsp = nsp

    def get_pending(self, node: expr) -> PendingExprGeneric:
        if isinstance(node, NamedExpr):
            return PendingNamedExpr(node, self.nsp)
        elif isinstance(node, Name):
            return PendingName(node, self.nsp)
        elif isinstance(node, (ListComp, SetComp, DictComp, GeneratorExp)):
            return PendingComp(node, self.nsp)
        else:
            return PendingExprGeneric(node)

    def cvt(self, node: expr):
        unconverted = node
        converted = None
        while True:
            pending_node = self.get_pending(unconverted)
            self.pending_stack.append(pending_node)
            unconverted = None
            while unconverted is None:
                try:
                    unconverted = self.pending_stack[-1].iter_fields.send(converted)
                    converted = None
                except StopIteration:
                    converted = self.pending_stack.pop().get_result()
                    if len(self.pending_stack) == 0:
                        return converted


def expr_transf(nsp: Namespace, node: expr):
    return ExpressionTransformer(nsp).cvt(node)
